
-- Create Date:    14:59:12 06/27/2012 

library IEEE;

use IEEE.STD_LOGIC_1164.ALL;
use IEEE.NUMERIC_STD.ALL;

entity fifo_window is
    generic(
        B: natural := 8;      -- Number of bits 
        W: natural := 3       -- Number of adress bits / window's length
     ); 
    Port ( clk     : in  STD_LOGIC;
           reset   : in  STD_LOGIC;
           rd      : in  STD_LOGIC;
           wr      : in  STD_LOGIC;
           wr_data : in  UNSIGNED (B-1 downto 0);
           rd_data : out  UNSIGNED (B-1 downto 0);
           empty   : out  STD_LOGIC;
           full    : out  STD_LOGIC);
end fifo_window;

architecture Behavioral of fifo_window is

    -- Register file definition 
    type reg_file_type is array (2**w -1 downto 0) of UNSIGNED(B-1 downto 0); 
    signal array_reg : reg_file_type; 

    -- FSMD signals
    signal w_ptr_reg, w_ptr_next, w_ptr_succ : std_logic_vector(w-1 downto 0); 
    signal r_ptr_reg, r_ptr_next, r_ptr_succ : std_logic_vector(w-1 downto 0);
    signal empty_reg, empty_next             : std_logic; 
    signal full_reg, full_next               : std_logic; 

    -- 2 bits represent 4 operations: 
    -- nothing 00
    -- only read 01
    -- only write 10
    -- read & write 11
    signal wr_op: std_logic_vector(1 downto 0);  
    signal wr_en: std_logic;

begin

wr_op  <= wr & rd; 
wr_en  <= wr and (not full_reg);

--------------------------------------------------------------------
---      register file (write port)
--------------------------------------------------------------------
 
process(clk, reset)
begin
    if rising_edge(clk) then
        -- write port
        if wr_en = '1' then
            array_reg (to_integer(unsigned(w_ptr_reg))) <= wr_data; 
        end if;
        
        -- read port
        rd_data <= array_reg (to_integer(unsigned(r_ptr_reg)));
    end if;
end process;

--------------------------------------------------------------------
--      FIFO Control logic
--------------------------------------------------------------------

-- register for read and write pointers
process(clk, reset)
begin
    if rising_edge(clk) then
        if reset = '1' then
            w_ptr_reg <= (others => '0');  
            r_ptr_reg <= (others => '0');
            full_reg  <= '0'; 
            empty_reg <= '1';
            
        else
            w_ptr_reg <= w_ptr_next; 
            r_ptr_reg <= r_ptr_next; 
            full_reg  <= full_next; 
            empty_reg <= empty_next; 
        end if;
    end if;
end process; 

-- succesive pointer values
w_ptr_succ <= std_logic_vector(unsigned(w_ptr_reg)+1); 
r_ptr_succ <= std_logic_vector(unsigned(r_ptr_reg)+1); 

-- next state logic for read and write pointers
process(w_ptr_reg, w_ptr_succ, r_ptr_reg, r_ptr_succ, full_reg, 
        empty_reg, wr_op)
begin
    w_ptr_next <= w_ptr_reg; 
    r_ptr_next <= r_ptr_reg; 
    full_next  <= full_reg; 
    empty_next <= empty_reg; 
    
    case wr_op is
        when "00" =>   -- nothing
        
        when "01" =>   -- only read
            if(empty_reg /= '1') then
                r_ptr_next <= r_ptr_succ; 
                full_next <= '0'; 
                if (r_ptr_succ = w_ptr_reg) then
                    empty_next <= '1'; 
                end if; 
            end if; 
        
        when "10" =>  -- only write
            if(full_reg /= '1') then
                w_ptr_next <= w_ptr_succ; 
                empty_next <= '0';
                if (w_ptr_succ = r_ptr_reg) then
                    full_next <= '1'; 
                end if; 
            end if; 
            
        when others => -- read & write
            w_ptr_next <= w_ptr_succ; 
            r_ptr_next <= r_ptr_succ;
    end case; 
    
end process; 

-- Output 
full  <= full_reg; 
empty <= empty_reg; 


end Behavioral;

