library ieee;
use ieee.std_logic_1164.all;
use ieee.numeric_std.all;

entity uart_tx is
    generic (
        B  : integer :=  8;
        s_ticks : integer := 16
    );
    port (
        clk       : in  std_logic;
        reset     : in  std_logic;
        baud_tick : in  std_logic;
        tx_start  : in  std_logic;
        dinp      : in  std_logic_vector (B-1 downto 0);
        tx        : out std_logic;
        done_tick : out std_logic
    );
end uart_tx;

architecture behavioral of uart_tx is
    type state_type is (idle, read_data, start, data, stop);
    signal state_cur, state_next     : state_type;
    signal s_cur, s_next             : unsigned(3 downto 0);
    signal n_cur, n_next             : unsigned(2 downto 0);
    signal tx_cur, tx_next           : std_logic;
    signal dbuffer_cur, dbuffer_next : std_logic_vector(B-1 downto 0);
begin
    -- register
    process(clk, reset)
    begin
        if rising_edge(clk) then
            if reset = '1' then
                state_cur    <= idle;
                s_cur        <= (others => '0');
                n_cur        <= (others => '0');
                dbuffer_cur  <= (others => '0');
                tx_cur       <= '1';
            else
                state_cur    <= state_next;
                s_cur        <= s_next;
                n_cur        <= n_next;
                tx_cur       <= tx_next;
                dbuffer_cur  <= dbuffer_next;
            end if;
        end if;
    end process;

    -- next state logic
    process(baud_tick, tx_start, dinp, dbuffer_cur, state_cur, s_cur, n_cur, tx_cur)
    begin
        state_next   <= state_cur;
        s_next       <= s_cur;
        n_next       <= n_cur;
        tx_next      <= tx_cur;
        dbuffer_next <= dbuffer_cur;
        done_tick  <= '0';

        case state_cur is
            when idle =>
                tx_next <= '1';

                if tx_start = '1' then
                    state_next   <= read_data;
                    s_next       <= (others => '0');
                end if;
                
            when read_data =>
                state_next   <= start;
                dbuffer_next <= dinp;
                
            when start =>
                tx_next <= '0';

                if baud_tick = '1' then
                    if s_cur = 15 then
                        state_next <= data;
                        s_next     <= (others => '0');
                        n_next     <= (others => '0');
                    else
                        s_next <= s_cur + 1;
                    end if;
                end if;

            when data => -- verify and then invert
                tx_next <= dbuffer_cur(to_integer(n_cur));

                if baud_tick = '1' then
                    if s_cur = 15 then
                        s_next <= (others => '0');

                        if n_cur = B - 1 then
                            state_next <= stop;
                        else
                            n_next <= n_cur + 1;
                        end if;
                    else
                        s_next <= s_cur + 1;
                    end if;
                end if;

            when stop =>
                tx_next <= '1';

                if baud_tick = '1' then
                    if s_cur = s_ticks - 1 then
                        state_next <= idle;
                        done_tick  <= '1';
                    else
                        s_next <= s_cur + 1;
                    end if;
                end if;
        end case;

    end process;

    -- output logic
    tx <= tx_cur;

end behavioral;